// -*- C++ -*-
//
// Package:     PhysicsToolsObjects
// Class  :     MVAComputer
// 

// Implementation:
//     getProcessor() and addProcessor() methods to add processors to
//     the discriminator computer calibration object. POOL doesn't support
//     polymorph pointers, so this is implemented using multiple containers
//     for each possible sub-class and an index array from which the
//     array of pointers can be reconstructed. In order to avoid having
//     to handle each sub-class container by hand here, the generated
//     reflex dictionary is used to find and read/write the vector<...>
//     containers for the individual classes in the private data members.
//     So changes can be solely done in the header files and does not leave
//     a trail elsewhere.
//
// Author:      Christophe Saout
// Created:     Sat Apr 24 15:18 CEST 2007
// $Id: MVAComputer.cc,v 1.2 2007/05/10 13:42:41 saout Exp $
//
#include <functional>
#include <algorithm>
#include <typeinfo>
#include <iostream>
#include <cstring>
#include <cstddef>
#include <string>
#include <vector>
#include <map>

// #include <Reflex/Reflex.h>

#include "FWCore/Utilities/interface/Exception.h"

#include "CondFormats/PhysicsToolsObjects/interface/MVAComputer.h"

using namespace std;

namespace PhysicsTools {
namespace Calibration {

string VarProcessor::getInstanceName() const
{
  return "VarProcessor";
  /*
  static const char prefix[] = "PhysicsTools::Calibration::";
  string type = ROOT::Reflex::Tools::Demangle(typeid(*this));
  if (type.size() <= sizeof prefix - 1 ||
      type.substr(0, sizeof prefix - 1) != prefix)
    throw cms::Exception("MVAComputerCalibration")
      << "getInstanceName failed for "
      << typeid(*this).name() << "." << endl;

  cout << "[Calibration::VarProcessor] ::getInstanceName=" << type.substr(sizeof prefix - 1);
  return type.substr(sizeof prefix - 1);
  */
}

MVAComputer::~MVAComputer()
{
  for ( vector< VarProcessor * >::const_iterator i=processors_.begin(); 
        i!=processors_.end() ; ++i )
    delete (*i);
}

MVAComputer::MVAComputer() {}

vector<VarProcessor*> MVAComputer::getProcessors() const
{
  return processors_;
}

void MVAComputer::addProcessor(const VarProcessor *proc)
{
  processors_.push_back ( new VarProcessor ( *proc ) );
}

MVAComputer &MVAComputerContainer::add(const string &label)
{
  entries.push_back(make_pair(label, MVAComputer()));
  return entries.back().second;
}

namespace {
  struct Comparator :
    public unary_function<const string&, bool> {

    inline Comparator(const string &label) : label(label) {}

    inline bool
    operator () (const MVAComputerContainer::Entry &entry) const
    { return entry.first == label; }

    const string &label;
  };
}

const MVAComputer &MVAComputerContainer::find(const string &label) const
{
  vector<Entry>::const_iterator pos =
        find_if(entries.begin(), entries.end(),
                     Comparator(label));

  if (pos == entries.end())
    throw cms::Exception("MVAComputerCalibration")
      << "Calibration record " << label
      << " not found in MVAComputerContainer." << endl;

  return pos->second;
}

} // namespace Calibration
} // namespace PhysicsTools
