//
// C++ Interface: KinematicTreeFactory
//
// Description:
//
//
// Author: Fabian Moser <fabiamos@gmail.com>, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef RAVEKINEMATICTREEFACTORY_H
#define RAVEKINEMATICTREEFACTORY_H

#include <rave/KinematicParticle.h>
#include <rave/KinematicTree.h>
#include <rave/ConstantMagneticField.h>

namespace rave
{

/**
 * @class KinematicTreeFactory provides the interface for fitting a decay chain
 * from given particles.
 *
 * Together with KinematicParticleFactory, KinematicTreeFactory  is at the
 * heart of the Rave kinematic fit capabilities. While the first creates
 * particle representations out of track data, the second fits these particle
 * representations called KinematicParticles to a decay chain represented by
 * a KinematicTree instance.
 *
 * @author Fabian Moser <fabiamos@gmail.com>
 */
class KinematicTreeFactory
{
  public:
    KinematicTreeFactory ( const rave::MagneticField & field = ConstantMagneticField(),
                           const std::string & method = "default",
                           int verbosity = 1 );

    ~KinematicTreeFactory();

    /**
     * Uses KinematicParticleVertexFitter
     * @param particles Field of particles
     * @return The fitted tree
     */
    rave::KinematicTree useVertexFitter (
      std::vector< rave::KinematicParticle > particles ) const;

    /**
     * Does a constraint kinematic vertex fit using the KinematicConstrainedVertexFitter
     * @param particles Field of particles
     * @param constr Constraint
     * @param fnd String selecting the used LinearisationPointFinder one of
     * lms,
     * @return The fitted tree
     */
    rave::KinematicTree useVertexFitter (
      std::vector< rave::KinematicParticle > particles,
      rave::KinematicConstraint cs,
      std::string fnd = "lms" ) const;

    /**
     * Uses KinematicParticleFitter
     * @param trees
     * @param cs
     * @param fu String selecting the fitter-updator pair. This is formatted
     * as fitter:updator where fitter is one of (lagrange) and updator is
     * one of (lagrange)
     * @return The refitted trees
     */
    std::vector< rave::KinematicTree > useParticleFitter (
      std::vector< rave::KinematicTree > trees,
      rave::KinematicConstraint cs,
      std::string fu = "lagrange:lagrange" );

  private:
    std::string theMethod;
    rave::MagneticField * theField;
    int theVerbosity;
};

}

#endif
