#ifndef _RaveVertex_H_
#define _RaveVertex_H_

#include <rave/Track.h>
#include <rave/Point3D.h>
#include <rave/Covariance3D.h>
#include <rave/BasicVertex.h>
#include <utility>
#include <vector>

namespace rave 
{

class Vertex : private BasicVertex::Proxy
{
  typedef BasicVertex::Proxy Base;
  /**
   *  The output of the rave library: a reconstructed vertex.
   *  A position in 3d space, plus its error.
   *  Along with the (refitted!!) tracks and
   *  the association probabilities of tracks belonging to this vertex.
   *  also a (fractional) number of degrees of freedom (ndf)
   *  and chi-squared.
   */
  public:
    Vertex(); //< default constructor, creating invalid vertex. 
    Vertex( const BasicVertex & );
    
    const Point3D & position() const; //< The vertex position
    const Covariance3D & error() const; //< The position's error

    /**
     *  the tracks plus the track weights
     */
    const std::vector < std::pair < double, Track > > & weightedTracks() const;

    std::vector < Track > tracks() const; //< the original tracks
    std::vector < Track > refittedTracks() const; //< the refitted tracks
    /**
     *  the refitted tracks, with the weights
     */
    const std::vector < std::pair < double, Track > > & weightedRefittedTracks() const;
    double ndf() const; //< number of degrees of freedom (may be fractional)
    double chiSquared() const; //< the total vertex fit chi squared.
    int id() const; // the vertex "id"
    bool isValid() const;
    bool hasRefittedTracks() const; //< do we have a collection of smoothed tracks?
    bool operator< ( const rave::Vertex & ) const;
    bool operator== ( const rave::Vertex & ) const;

    /**
     *  The vertex components - only used in the Gaussian
     *  algorithm (not yet available)
     */
    std::vector < Vertex > components() const;

    Vertex( const Point3D &, const Covariance3D &,
            const std::vector < std::pair < double, Track > > & tracks, double ndf, double chi2 );
    Vertex( const Point3D &, const Covariance3D &,
            const std::vector < Track > & tracks, double ndf, double chi2 );
    Vertex( const Point3D &, const Covariance3D &,
            const std::vector < std::pair < double, Track > > & tracks, 
            const std::vector < std::pair < double, Track > > & smoothedtracks, double ndf, double chi2 );
    Vertex( const Point3D &, const Covariance3D &,
            const std::vector < Track > & tracks,
            const std::vector < Track > & smoothedtracks, double ndf, double chi2 );
};

}

#endif
