#include "RaveBase/RaveInterface/rave/KinematicTreeFactory.h"

#include "FWCore/MessageLogger/interface/MessageLogger.h"
#include "RecoVertex/KinematicFit/interface/KinematicParticleVertexFitter.h"
#include "RecoVertex/KinematicFit/interface/KinematicConstrainedVertexFitter.h"
#include "RecoVertex/KinematicFit/interface/KinematicParticleFitter.h"
#include "RecoVertex/KinematicFit/interface/LagrangeParentParticleFitter.h"
#include "RecoVertex/KinematicFit/interface/LagrangeChildUpdator.h"
#include "RecoVertex/KinematicFitPrimitives/interface/KinematicConstraint.h"
#include "RecoVertex/KinematicFitPrimitives/interface/MultiTrackKinematicConstraint.h"

#include "RaveBase/Converters/interface/MagneticFieldWrapper.h"
#include "RaveTools/Converters/interface/MagneticFieldSingleton.h"

#include "boost/bind.hpp"

namespace rave
{

KinematicTreeFactory::KinematicTreeFactory (
  const rave::MagneticField & field, const std::string & method,
  int verbosity ) :
    theField ( field.copy() ), theMethod ( method ), theVerbosity ( verbosity )
{
  MagneticFieldSingleton::Instance()->registry ( new MagneticFieldWrapper ( *theField ) );

  edm::LogInfo ( "rave::KinematicTreeFactory" )
  << " magnetic field at (0,0,0) is "
  << MagneticFieldSingleton::Instance()->inTesla ( GlobalPoint ( 0., 0., 0. ) )
  << endl;
}

KinematicTreeFactory::~KinematicTreeFactory()
{}

namespace
{
::RefCountedKinematicParticle doParticleCasting ( const boost::any & p )
{
  return boost::any_cast< ::RefCountedKinematicParticle > ( p );
}

::RefCountedKinematicTree doTreeCasting ( const boost::any & t )
{
  return boost::any_cast< ::RefCountedKinematicTree > ( t );
}
}

KinematicTree KinematicTreeFactory::useVertexFitter ( std::vector< KinematicParticle > particles ) const
{
  // The fitter we are going to use
  ::KinematicParticleVertexFitter fitter;

  // Create the input data field
  std::vector< ::RefCountedKinematicParticle > internalParticles;
  std::transform (
    particles.begin(),
    particles.end(),
    std::back_inserter ( internalParticles ),
    boost::bind ( &doParticleCasting,
                  boost::bind ( &BasicKinematicParticle::getInternal,
                                boost::bind ( &KinematicParticle::data, _1 ) ) ) );

  // Do the fitting, this is where the real work is done
  ::RefCountedKinematicTree internalTree = fitter.fit ( internalParticles );

  // Now parse the tree into the interface format
  return rave::KinematicTree ( BasicKinematicTree ( internalTree, particles ) );
}

rave::KinematicTree KinematicTreeFactory::useVertexFitter (
  std::vector< rave::KinematicParticle > particles,
  rave::KinematicConstraint cs, std::string fnd ) const
{
  // The fitter we are going to use
  ::KinematicConstrainedVertexFitter fitter;

  // Create the input data field
  std::vector< ::RefCountedKinematicParticle > internalParticles;
  std::transform (
    particles.begin(),
    particles.end(),
    std::back_inserter ( internalParticles ),
    boost::bind ( &doParticleCasting,
                  boost::bind ( &BasicKinematicParticle::getInternal,
                                boost::bind ( &KinematicParticle::data, _1 ) ) ) );

  // Unwrap the constraint
  MultiTrackKinematicConstraint * internalConstraint =
    boost::any_cast< boost::shared_ptr< ::MultiTrackKinematicConstraint > > (
      cs.data().getInternal() ).get();

  // Do the fitting, this is where the real work is done
  ::RefCountedKinematicTree internalTree = fitter.fit ( internalParticles, internalConstraint );

  // Now parse the tree into the interface format
  return rave::KinematicTree ( BasicKinematicTree ( internalTree, particles ) );
}

/// TODO Allow other fitter-updator pairs
std::vector< rave::KinematicTree > KinematicTreeFactory::useParticleFitter (
  std::vector< rave::KinematicTree > trees,
  rave::KinematicConstraint cs, std::string fu )
{
  // The fitter we are going to use
  ::KinematicParticleFitter fitter =
    ::KinematicParticleFitter( LagrangeParentParticleFitter(), LagrangeChildUpdator() );

  // Create the input data field
  std::vector< ::RefCountedKinematicTree > internalInputTrees;
  std::transform (
    trees.begin(),
    trees.end(),
    std::back_inserter ( internalInputTrees ),
    boost::bind ( &doTreeCasting,
                  boost::bind ( &BasicKinematicTree::getInternal,
                                boost::bind ( &KinematicTree::data, _1 ) ) ) );

  // Unwrap the constraint
  ::KinematicConstraint * internalConstraint =
    boost::any_cast< boost::shared_ptr< ::KinematicConstraint > > (
      cs.data().getInternal() ).get();

  // Do the fitting, this is where the real work is done
  std::vector< ::RefCountedKinematicTree > internalOutputTrees =
    fitter.fit ( internalConstraint, internalInputTrees );

  // Create the output data field
  std::vector< rave::KinematicTree > outputTrees;
  for ( std::vector< ::RefCountedKinematicTree >::iterator i = internalOutputTrees.begin();
        i != internalOutputTrees.end();
        i++ )
  {
    outputTrees.push_back( rave::KinematicTree( rave::BasicKinematicTree( *i ) ) );
  }

  return outputTrees;
}

}
