#include "RaveBase/RaveInterface/rave/VertexFactory.h"
#include "TrackingTools/TransientTrack/interface/TransientTrack.h"
#include "RecoVertex/VertexPrimitives/interface/TransientVertex.h"
#include "RaveBase/Converters/interface/CmsToRaveObjects.h"
#include "RaveBase/Converters/interface/RaveToCmsObjects.h"
#include "RaveBase/Converters/interface/RaveStreamers.h"
#include "RaveBase/Converters/interface/RaveComputations.h"
#include "RaveBase/RaveEngine/interface/RaveReconstructor.h"
#include "RaveBase/Converters/interface/MagneticFieldWrapper.h"
#include "RaveTools/Converters/interface/MagneticFieldSingleton.h"
#include "FWCore/MessageLogger/interface/MessageLogger.h"
#include <iostream>
#include <iomanip>

using namespace rave;
using namespace std;

namespace {
  RaveReconstructor * getRector ( const std::string & method )
  {
    if ( method == "none" )
    {
      return (RaveReconstructor *) 0;
    }
    return new RaveReconstructor ( method );
  }
}

VertexFactory::VertexFactory ( const rave::MagneticField & field,
                   const rave::Propagator & prop,
                   const std::string & method,
                   int verbosity ) :
  theRector ( getRector ( method ) ), theMethod ( method ),
  theField ( field.copy() ) , theProp ( prop.copy() ),
  theBeamspot ( Ellipsoid3D() ), 
  theVerbosity ( verbosity )
  
{
  edm::LogInfo ( "VertexFactory" ) << " factory called w/o beamspot!";
  // createRecoBeamSpot();
  setup();
}

void VertexFactory::createRecoBeamSpot()
{
  math::XYZPoint p ( theBeamspot.point().x(), theBeamspot.point().y(), theBeamspot.point().z() );
  math::Error<7> error;
  float cxx=theBeamspot.covariance().dxx();
  float cyy=theBeamspot.covariance().dyy();
  float czz=theBeamspot.covariance().dzz();
  error(0,0)=cxx;
  error(1,1)=cyy;
  error(2,2)=czz;
  error(1,0)=theBeamspot.covariance().dxy();
  error(2,0)=theBeamspot.covariance().dxz();
  error(2,1)=theBeamspot.covariance().dyz();
  reco::BeamSpot * tmp = new reco::BeamSpot ( p, sqrt ( czz ), 0., 0., 
      sqrt ( (cxx + cyy)/2. ) , error );
  theRecoBeamSpot = (void *) (tmp);
  if ( false )
  {
    edm::LogWarning("VertexFactory::createRecoBeamSpot") 
      << " created: "
      << " sxx[nm]=" << 10000000. * sqrt ( cxx )
      << " syy[nm]=" << 10000000. * sqrt ( cyy )
      << " szz[nm]=" << 10000000. * sqrt ( czz );
    edm::LogWarning("VertexFactory::createRecoBeamSpot") 
      << " created: "
      << " txx[nm]=" << 10000000. * sqrt ( tmp->covariance(0,0) )
      << " tyy[nm]=" << 10000000. * sqrt ( tmp->covariance(1,1) )
      << " tzz[nm]=" << 10000000. * sqrt ( tmp->covariance(2,2) );
    edm::LogWarning("VertexFactory::createRecoBeamSpot") 
      << " rxx[nm]=" << 10000000. * sqrt ( tmp->rotatedCovariance3D()(0,0) )
      << " rxy[nm]=" << 10000000. * sqrt ( tmp->rotatedCovariance3D()(0,1) )
      << " ryy[nm]=" << 10000000. * sqrt ( tmp->rotatedCovariance3D()(1,1) )
      << " rzz[nm]=" << 10000000. * sqrt ( tmp->rotatedCovariance3D()(2,2) );
    edm::LogWarning("VertexFactory::createRecoBeamSpot") << "   `- sgm_z=" << tmp->sigmaZ() << ", " << tmp->covariance(2,2);
    edm::LogWarning("VertexFactory::createRecoBeamSpot") << "   `- beamwidth=" << tmp->BeamWidth();
    edm::LogWarning("VertexFactory::createRecoBeamSpot") << "   `- dxdz=" << tmp->dxdz();
    edm::LogWarning("VertexFactory::createRecoBeamSpot") << "   `- dydz=" << tmp->dydz();
  }
}

VertexFactory::VertexFactory ( const rave::MagneticField & field,
                   const rave::Propagator & prop,
                   const rave::Ellipsoid3D & beamspot,
                   const std::string & method,
                   int verbosity ) :
  theRector ( getRector ( method ) ), theMethod ( method ),
  theField ( field.copy() ) , theProp ( prop.copy() ),
  theBeamspot ( beamspot ), theRecoBeamSpot ( 0 ), theVerbosity ( verbosity )
{
  createRecoBeamSpot();  
  // cout << "[rave::VertexFactory] before setup 2" << endl;
  setup();
}

void VertexFactory::setup()
{
  // cout << "[rave::VertexFactory] debug: setup starts" << endl;
    MagneticFieldSingleton::Instance()->registry ( new MagneticFieldWrapper ( *theField ) );
    /*
    if (!theRector)
    {
      cout << "[rave::VertexFactory] error: could not initialize reconstructor: " << theMethod
           << endl;
    }*/

    if ( theVerbosity > 20 )
    {
      cout << "[rave::VertexFactory] magnetic field at (0,0,0) is "
           << MagneticFieldSingleton::Instance()->inTesla ( GlobalPoint (0.,0.,0.) )
           << endl;
    }
}

VertexFactory::~VertexFactory()
{
    // cout << "[VertexFactory] destor!" << endl;
    if ( theField ) delete theField;
    if ( theProp) delete theProp;
    if ( theRector ) delete theRector;
    for ( map < string, RaveReconstructor * >::const_iterator i=theRectors.begin(); 
          i!=theRectors.end() ; ++i )
    {
      delete i->second;
    }
    theRectors.clear();
}

vector < Vertex > VertexFactory::create ( const vector < Track > & trks, bool use_bs ) const
{
    if ( !theRector )
    {
      cout << "[VertexFactory] error: trying to use uninitialised factory!"
           << endl;
      return vector < Vertex > ();
    }
    return fit ( trks, *theRector, use_bs );
}

vector < Vertex > VertexFactory::create ( const vector < Track > & trks, const string & method, bool use_bs ) const
{
  RaveReconstructor * tmp = theRectors[method];
  if ( !tmp )
  {
    tmp = getRector ( method );
    if (!tmp)
    {
      cout << "[VertexFactory] error cannot initialse \"" << method << "\"" << endl;
      return vector < Vertex > ();
    }
    theRectors[method] = tmp;
  }
  return fit ( trks, *(tmp), use_bs );
}

VertexFactory VertexFactory::operator= ( const VertexFactory & o )
{
  // cout << "[rave::VertexFactory] assignment op! theProp=" << o.theProp << endl;
  if ( o.theProp && o.theField )
  {
    return VertexFactory ( *(o.theField), *(o.theProp), o.theMethod, o.theVerbosity );
  }
  return VertexFactory();
}

VertexFactory::VertexFactory ( const VertexFactory & o ) :
   theRector ( 0 ), theMethod ( o.theMethod ), theField ( 0 ),
   theProp ( 0 ), theBeamspot ( o.theBeamspot ), 
   theRecoBeamSpot ( 0 ), theVerbosity ( o.theVerbosity )
{
  cout << "[rave::VertexFactory] copy constructor." << endl;
  if ( o.theRector ) theRector = o.theRector->clone();
  if ( o.theField ) theField = o.theField->copy();
  if ( o.theProp ) theProp = o.theProp->copy();
  for ( map < string, RaveReconstructor *  >::const_iterator i=o.theRectors.begin();
        i!=theRectors.end() ; ++i )
  {
    theRectors[i->first]=i->second->clone();
  }
}

vector < Vertex > VertexFactory::fit ( const vector < Track > & trks, const RaveReconstructor & r,
   bool use_bs ) const
{
    vector < rave::Vertex > ret;
    vector < reco::TransientTrack > ttrks;
    RaveToCmsObjects fconverter;
    for ( vector< Track >::const_iterator i=trks.begin(); i!=trks.end() ; ++i )
    {
        ttrks.push_back ( fconverter.tTrack ( *i ) );
    }

    vector < TransientVertex > tmp;
    // LogDebug("") << "use_bs=" << use_bs;
    if ( use_bs && hasBeamspot () )
    {
      reco::BeamSpot * bs = ( reco::BeamSpot * ) (theRecoBeamSpot);
      // LogDebug("") << "   bs=" << (void *) bs;
      if ( theRecoBeamSpot == 0 )
      {
        edm::LogError("") << "ouch! theRecoBeamSpot=0!!";
        tmp=r.vertices ( ttrks );
      } else {
        tmp = r.vertices ( ttrks, *bs );
      }
    } else {
      tmp = r.vertices ( ttrks );
    }
    
    CmsToRaveObjects converter;
    
    for ( vector< TransientVertex >::const_iterator i=tmp.begin(); 
          i!=tmp.end() ; ++i )
    {
      ret.push_back ( converter.convert (*i, trks ) );
    }

    if ( theVerbosity>1 )
    {
        cout << "[rave::VertexFactory] produced " << ret.size() << " vertices from " << ttrks.size()
                  << " tracks." << endl;
    };
    return ret;
}

const int VertexFactory::verbosity() const
{
  return theVerbosity;
}

const std::string & VertexFactory::method() const
{
  return theMethod;
}

bool VertexFactory::hasBeamspot() const
{
  return theBeamspot.isValid();
}

const rave::Ellipsoid3D & VertexFactory::getBeamspot() const
{
  return theBeamspot;
}

const rave::Propagator & VertexFactory::getPropagator() const
{
    return *theProp;
}

const rave::MagneticField & VertexFactory::getMagneticField() const
{
    return *theField;
}

const RaveReconstructor & VertexFactory::reconstructor() const
{
  return (*theRector);
}

void VertexFactory::setDefaultMethod ( const string & method )
{
  if ( theRector ) delete theRector;
  theRector = getRector ( method );
  theMethod=method;
}
