#include "TrackingTools/PatternTools/interface/ImpactPointExtrapolator.h"
#include "TrackingTools/TrajectoryState/interface/TrajectoryStateOnSurface.h"
#include "TrackingTools/TrajectoryState/interface/FreeTrajectoryState.h"
#include "DataFormats/TrajectorySeed/interface/PropagationDirection.h"
#include "DataFormats/GeometryVector/interface/GlobalVector.h"
#include "TrackingTools/GeomPropagators/interface/AnalyticalPropagator.h"
#include "DataFormats/GeometrySurface/interface/PlaneBuilder.h"
#include "RaveTools/Converters/interface/MagneticFieldSingleton.h"

ImpactPointExtrapolator::ImpactPointExtrapolator () :
  thePropagator(new AnalyticalPropagator(MagneticFieldSingleton::Instance(),anyDirection)) {}

ImpactPointExtrapolator::ImpactPointExtrapolator (const Propagator& u) :
  thePropagator(u.clone()) 
{
  thePropagator->setPropagationDirection(anyDirection);
}

TrajectoryStateOnSurface ImpactPointExtrapolator::extrapolate(
  const FreeTrajectoryState& fts, const GlobalPoint& vtx) const
{
  // use default BidirectionalPropagator
  return extrapolateSingleState(fts, vtx, *thePropagator);
}

TrajectoryStateOnSurface ImpactPointExtrapolator::extrapolate(
  const TrajectoryStateOnSurface tsos, const GlobalPoint& vtx) const
{
  // use default BidirectionalPropagator
  if ( tsos.isValid() )  return extrapolateFullState(tsos, vtx, *thePropagator);
  else  return tsos;
}

 TrajectoryStateOnSurface ImpactPointExtrapolator::extrapolate(
  const FreeTrajectoryState& fts, const GlobalPoint& vtx, 
  const Propagator& u) const
{
  // construct BidirectionalPropagator
  DeepCopyPointerByClone<Propagator> p(u.clone());
  p->setPropagationDirection(anyDirection);
  return extrapolateSingleState(fts,vtx,*p);
} 

 TrajectoryStateOnSurface ImpactPointExtrapolator::extrapolate(
  const TrajectoryStateOnSurface tsos, const GlobalPoint& vtx, 
  const Propagator& u) const
{
  if ( tsos.isValid() ) {
    // construct BidirectionalPropagator
    DeepCopyPointerByClone<Propagator> p(u.clone());
    p->setPropagationDirection(anyDirection);
    return extrapolateFullState(tsos,vtx,*p);
  }
  return tsos;
} 

TrajectoryStateOnSurface 
ImpactPointExtrapolator::extrapolateFullState (const TrajectoryStateOnSurface tsos, 
					       const GlobalPoint& vtx, 
					       const Propagator& p) const
{
  //
  // first determine IP plane using propagation with (single) FTS
  // could be optimised (will propagate errors even if duplicated below)
  //
  TrajectoryStateOnSurface singleState = 
    extrapolateSingleState(*tsos.freeTrajectoryState(),vtx,p);
  if ( !singleState.isValid() || tsos.components().size()==1 )  return singleState;
  //
  // propagate multiTsos to plane found before
  //
  return p.propagate(tsos,singleState.surface());
}

TrajectoryStateOnSurface 
ImpactPointExtrapolator::extrapolateSingleState (const FreeTrajectoryState& fts, 
						 const GlobalPoint& vtx, 
						 const Propagator& p) const
{
  // origin of target surface
  Surface::PositionType pos(vtx);

  // create a fts without errors for faster propagation
  FreeTrajectoryState fastFts(fts.parameters());
  GlobalVector dir = fastFts.momentum().unit();

  bool refining = true;
  int iterations = 100;
  PlaneBuilder pBuilder;
  while (refining && iterations > 0) {
    iterations--;

    // describe orientation of target surface on basis of track parameters
    double wx = dir.x();
    double wy = dir.y();
    double wz = dir.z();
    double dydx = wy / wx;
    double uy = 1. / sqrt(1 + dydx * dydx);
    double ux = - dydx * uy;
    Surface::RotationType rot(    ux,     uy,              0,
			      -wz*uy,  wz*ux,  wx*uy - wy*ux,
				  wx,     wy,             wz);
    ReferenceCountingPointer<BoundPlane> bp = pBuilder.plane(pos, rot);
    
    // extrapolate fastFts to target surface
    TrajectoryStateOnSurface tsos = p.propagate(fastFts, *bp);
    
    if (!tsos.isValid()) {
      return tsos;
    } else {
      GlobalVector oldDir = dir;
      dir = tsos.globalMomentum().unit();
      if (fabs(oldDir.dot(dir)) > 0.99999) refining = false;
    }

    // invalid TrajectoryStateOnSurface to be returned in case of trouble
    if (iterations == 0)  return TrajectoryStateOnSurface();
  }

  // now propagate with errors
  double wx = dir.x();
  double wy = dir.y();
  double wz = dir.z();
  double dydx = wy / wx;
  double uy = 1. / sqrt(1 + dydx * dydx);
  double ux = - dydx * uy;
  Surface::RotationType rot(    ux,     uy,              0,
			    -wz*uy,  wz*ux,  wx*uy - wy*ux,
				wx,     wy,             wz);
  ReferenceCountingPointer<BoundPlane> bPlane = pBuilder.plane(pos, rot);
  TrajectoryStateOnSurface tsos = p.propagate(fts, *bPlane);
  if (!tsos.isValid()) {
    return tsos;
  }

  // rotation to have x-axis of returned surface parallel to impact vector
  GlobalVector ip = tsos.globalPosition() - vtx;
  ip = ip.unit();
  ux = ip.x();
  uy = ip.y();
  double uz = ip.z();
  Surface::RotationType ipRot(          ux,           uy,           uz,
			       wy*uz-wz*uy,  wz*ux-wx*uz,  wx*uy-wy*ux,
				        wx,           wy,           wz);
//    Surface* ipSurf = new BoundPlane(pos, ipRot);
  ReferenceCountingPointer<BoundPlane> ipSurf = pBuilder.plane(pos, ipRot);

  return TrajectoryStateOnSurface( *(tsos.freeTrajectoryState()), 
				   *ipSurf);
}
